__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

import numpy as np

from jina.executors.decorators import batching
from jina.executors.encoders import BaseNumericEncoder


class TSNEEncoder(BaseNumericEncoder):
    """
    Encode data using t-distributed Stochastic Neighbor Embedding.

    Encodes data from an ndarray in size `B x T` into an ndarray in size `B x D`
    Where `B` is the batch's size and `T` and `D` are the dimensions pre (`T`)
    and after (`D`) the encoding.

    See more details
    `here <https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html>`_

    :param output_dim: Dimension of the embedded space
    :param random_state: Used to seed the cost_function of TSNE
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments

    .. note:
        Unlike other numeric encoders, TSNE does not inherit Transform encoder
        because it can't have a transform without fit.
    """

    def __init__(self, output_dim: int = 64,
                 random_state: int = 2020,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_dim = output_dim
        self.random_state = random_state

    def post_init(self):
        """Load TSNE model"""
        super().post_init()
        from sklearn.manifold import TSNE
        self.model = TSNE(n_components=self.output_dim, random_state=self.random_state)

    @batching
    def encode(self, data: 'np.ndarray', *args, **kwargs) -> 'np.ndarray':
        """
        Encode data from an ndarray in size `B x T` into an ndarray in size `B x D`

        :param data: a `B x T` numpy ``ndarray``, `B` is the size of the batch
        :return: a `B x D` numpy ``ndarray``
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        """
        return self.model.fit_transform(data)



__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Union, Tuple

from .. import QuerySetReader, FlatRecursiveMixin, BaseRecursiveDriver

# noinspection PyUnreachableCode
if False:
    from ...types.sets import DocumentSet


class ExcludeQL(QuerySetReader, FlatRecursiveMixin, BaseRecursiveDriver):
    """Clean some fields from the document-level protobuf to reduce the total size of the request
    Example::
    - !ExcludeQL
    with:
        fields:
            - chunks
            - buffer

    ExcludeQL will avoid `buffer` and `chunks` fields to be sent to the next `Pod`

    :param fields: the pruned field names in tuple
    :param traversal_paths: the traversal paths
    :param *args: *args
    :param **kwargs: **kwargs
    """

    def __init__(
        self,
        fields: Union[Tuple, str],
        traversal_paths: Tuple[str] = ('r',),
        *args,
        **kwargs
    ):
        super().__init__(traversal_paths=traversal_paths, *args, **kwargs)
        if isinstance(fields, str):
            self._fields = [fields]
        else:
            self._fields = [field for field in fields]

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs):
        for doc in docs:
            for k in self.fields:
                doc.ClearField(k)


class SelectQL(ExcludeQL):
    """Selects some fields from the chunk-level protobuf to reduce the total size of the request, it works with the opposite
    logic as `:class:`ExcludeQL`

        Example::
        - !SelectQL
        with:
            fields:
                - matches

        SelectQL will ensure that the `outgoing` documents only contain the field `matches`
    """

    def _apply_all(self, docs: 'DocumentSet', *args, **kwargs):
        for doc in docs:
            for k in doc.DESCRIPTOR.fields_by_name.keys():
                if k not in self.fields:
                    doc.ClearField(k)


class ExcludeReqQL(ExcludeQL):
    """Clean up request from the request-level protobuf message to reduce the total size of the message

    This is often useful when the proceeding Pods require only a signal, not the full message.
    """

    def __call__(self, *args, **kwargs):
        """


        .. # noqa: DAR102


        .. # noqa: DAR101
        """
        for k in self.fields:
            self.req.ClearField(k)


class SelectReqQL(ExcludeReqQL):
    """Clean up request from the request-level protobuf message to reduce the total size of the message, it works with the opposite
    logic as `:class:`ExcludeReqQL`


    .. # noqa: DAR101
    """

    def __call__(self, *args, **kwargs):
        """


        .. # noqa: DAR102


        .. # noqa: DAR101
        """
        for k in self.req.DESCRIPTOR.fields_by_name.keys():
            if k not in self.fields:
                self.req.ClearField(k)



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict

import numpy as np
from jina.executors.crafters import BaseCrafter


class AudioMonophoner(BaseCrafter):
    """:class:`AudioMonophoner` makes the audio signal monophonic on doc-level."""

    def craft(self, blob: np.ndarray, *args, **kwargs) -> Dict:
        """
        Read the `ndarray` of the audio signal.

        Makes the audio signal monophonic and saves the `ndarray` of the
        monophonic signal in the `blob` of the Document.

        :param blob: the ndarray of the audio signal
        :return: a Document dict with the monophonic audio signal
        :param args:  Additional positional arguments
        :param kwargs: Additional keyword arguments
        :return: A dictionary with the monophonic signal

        """
        import librosa
        signal_mono = librosa.to_mono(blob)

        return dict(offset=0, blob=signal_mono)




__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from jina.executors.encoders.numeric import TransformEncoder

class IncrementalPCAEncoder(TransformEncoder):
    """
    Encodes data from an ndarray in size `B x T` into an ndarray in size `B x D`.

    Where `B` is the batch's size and `T` and `D` are the dimensions pre (`T`)
    and after (`D`) the encoding.

    :param output_dim: the output size.
    :param num_features: the number of input features.
        If ``num_features`` is None, then ``num_features``
        is inferred from the data
    :param whiten: If whiten is false, the data is already considered to be whitened,
        and no whitening is performed.
    :param args:  Additional positional arguments
    :param kwargs: Additional keyword arguments

    More details can be found
    `here <https://scikit-learn.org/stable/auto_examples/decomposition/plot_incremental_pca.html>`_

    .. note::
        :class:`IncrementalPCAEncoder` must be trained before calling ``encode()``.
        This encoder can be trained in an incremental way.
    """

    def __init__(self, num_features: int = None, whiten: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.whiten = whiten
        self.num_features = num_features
        self.is_trained = False
        self.model = None

    def post_init(self):
        """Load IncrementalPCA model"""
        super().post_init()
        if not self.model:
            from sklearn.decomposition import IncrementalPCA
            self.model = IncrementalPCA(
                n_components=self.output_dim,
                whiten=self.whiten)


__copyright__ = "Copyright (c) 2020 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from pkg_resources import resource_filename

from .logger import JinaLogger

__all__ = ['default_logger', 'profile_logger']

default_logger = JinaLogger('JINA')  #: a logger at the global-level
profile_logger = JinaLogger(
    'PROFILE',
    log_config=resource_filename(
        'jina', '/'.join(('resources', 'logging.profile.yml'))
    ),
)



__copyright__ = "Copyright (c) 2021 Jina AI Limited. All rights reserved."
__license__ = "Apache-2.0"

from typing import Dict

import numpy as np
from jina.executors.rankers import Match2DocRanker


class LevenshteinRanker(Match2DocRanker):
    """
    :class:`LevenshteinRanker` Computes the negative Levenshtein distance
        between a query and its matches. The distance is negative, in order to
        achieve a bigger=better result, sort in the respective driver.
    """

    required_keys = {"text"}

    def score(
        self, query_meta: Dict, old_match_scores: Dict, match_meta: Dict
    ) -> "np.ndarray":
        """
        Calculate the negative Levenshtein distance

        :param query_meta: A Dict of queries to score
        :param old_match_scores: Previously scored matches
        :param match_meta: A Dict of matches of given query
        :return: an `ndarray` of the size ``M x 2``.
            `M`` is the number of new scores.
            The columns correspond to the ``COL_MATCH_ID`` and ``COL_SCORE``.

        """
        from Levenshtein import distance

        new_scores = [
            (
                match_id,
                -distance(query_meta['text'], match_meta[match_id]['text']),
            )
            for match_id, old_score in old_match_scores.items()
        ]
        return np.array(
            new_scores,
            dtype=[(self.COL_MATCH_ID, np.object), (self.COL_SCORE, np.float64)],
        )
